use crate::instructions::get_instruction;
use crate::memory::{MemHandle, MemoryReader, MemoryWriter};
use std::fmt::Display;
use std::sync::mpsc::{channel, Receiver, Sender};
use std::thread::sleep;
use std::time::Duration;

#[derive(Clone, Copy)]
pub enum StatusFlag {
    Negative = 0b1000_0000,
    Overflow = 0b0100_0000,
    Brk = 0b0011_0000,
    //BrkIrq = 0b0010_0000,
    Decimal = 0b0000_1000,
    IrqDisable = 0b0000_0100,
    Zero = 0b000_0010,
    Carry = 0b0000_0001,
}

#[derive(Clone, Debug)]
pub struct CpuController(Sender<CpuControl>);

pub enum CpuControl {
    Irq,
    Nmi,
    Toggle,
    Data,
    Cycle,
}

impl CpuController {
    pub fn new(sender: Sender<CpuControl>) -> Self {
        Self(sender)
    }
    pub fn irq(&self) {
        self.0.send(CpuControl::Irq);
    }
    pub fn nmi(&self) {
        self.0.send(CpuControl::Nmi);
    }
    pub fn toggle(&self) {
        self.0.send(CpuControl::Toggle);
    }
    pub fn cycle(&self) {
        self.0.send(CpuControl::Cycle);
    }
    pub fn data(&self) {
        self.0.send(CpuControl::Data);
    }
}

#[derive(Debug)]
pub struct CpuReceiver(Receiver<CpuControl>);
impl CpuReceiver {
    pub fn new(receiver: Receiver<CpuControl>) -> Self {
        Self(receiver)
    }
}

#[derive(Debug)]
pub struct Cpu {
    pub a: u8,   // Accumulator Register
    pub x: u8,   // X Register
    pub y: u8,   // Y Register
    pub pc: u16, // Program Counter
    pub s: u8,   // Stack Pointer
    pub p: u8,   // Status Register
    pub irq: bool,
    pub nmi: bool,
    pub memory: MemHandle,
    pub pending_cycles: usize,
    receiver: Option<CpuReceiver>,
    stopped: bool,
    cycle: bool,
}

impl Display for Cpu {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{},{}", self.a, self.x)
    }
}

impl MemoryReader for Cpu {
    fn read(&self, address: u16) -> u8 {
        self.memory.read(address)
    }
}
impl MemoryWriter for Cpu {
    fn write(&self, address: u16, data: u8) {
        self.memory.write(address, data);
    }
}

impl Cpu {
    pub fn new(memory: MemHandle) -> Self {
        Cpu {
            a: 0x00,
            x: 0x00,
            y: 0x00,
            pc: 0x0000,
            s: 0xFF,
            p: 0b0010_0100,
            irq: false,
            nmi: false,
            receiver: None,
            memory,
            stopped: false,
            pending_cycles: 0,
            cycle: false, // cycle_count: 0,
                          // state_tx,
        }
    }
    pub fn with_receiver(mut self, receiver: CpuReceiver) -> Self {
        self.receiver = Some(receiver);
        self
    }
    pub fn new_with_control(memory: MemHandle) -> (Self, CpuController) {
        let (tx, rx) = channel::<CpuControl>();
        let controller = CpuController(tx);
        let receiver = CpuReceiver(rx);
        let cpu = Cpu::new(memory).with_receiver(receiver);
        (cpu, controller)
    }
    pub fn reset(&mut self) {
        let reset_vector_pointer = self.read_word(0xFFFC);
        self.pc = reset_vector_pointer;
        self.a = 0;
        self.pending_cycles = 0;
    }
    pub fn read_word(&self, address: u16) -> u16 {
        let low_byte = self.read(address);
        let high_byte = self.read(address.wrapping_add(0x1));
        (high_byte as u16) << 8 | (low_byte as u16)
    }

    fn stack_addr(&self) -> u16 {
        // Dunno if this is necessary, i just don't like adding the 0x0100 every time
        0x0100 + self.s as u16
    }

    pub fn push_stack(&mut self, data: u8) {
        self.s = self.s.wrapping_sub(0x1);
        self.write(self.stack_addr(), data);
    }

    pub fn push_stack_word(&mut self, address: u16) {
        self.s = self.s.wrapping_sub(0x1);
        self.write(self.stack_addr(), address.to_le_bytes()[1]); // Upper byte first
        self.s = self.s.wrapping_sub(0x1);
        self.write(self.stack_addr(), address.to_le_bytes()[0]); // Lower byte second
    }

    pub fn pop_stack(&mut self) -> u8 {
        let byte = self.read(self.stack_addr());
        self.s = self.s.wrapping_add(0x1);
        byte
    }

    pub fn pop_stack_word(&mut self) -> u16 {
        let low_byte = self.pop_stack();
        let high_byte = self.pop_stack();
        ((high_byte as u16) << 8) + low_byte as u16
    }

    pub fn set_flag(&mut self, flag: StatusFlag, value: bool) {
        self.p &= !(flag as u8);
        if value {
            self.p |= flag as u8
        }
    }

    pub fn get_flag(&self, flag: StatusFlag) -> bool {
        (self.p & flag as u8) > 0
    }

    pub fn is_negative(&self, value: u8) -> bool {
        value & 0b1000_0000 == 0b1000_0000
    }

    pub fn wait_for_interrupt(&self) {
        unimplemented!()
    }

    pub fn interrupt(&mut self) {
        self.irq = false;
        self.push_stack_word(self.pc);
        self.push_stack(self.p);
        self.set_flag(StatusFlag::IrqDisable, true);
        self.pc = self.read_word(0xFFFE);
    }

    pub fn toggle_stopped(&mut self) {
        self.stopped = !self.stopped;
    }

    fn receive_control(&mut self) {
        match &self.receiver {
            Some(receiver) => {
                let control = match receiver.0.try_recv() {
                    Ok(control) => control,
                    Err(_error) => return, // most of the time we won't have any impending control
                                           // messages, just return if that's the case
                };
                match control {
                    CpuControl::Nmi => self.nmi = true,
                    CpuControl::Irq => self.irq = true,
                    CpuControl::Toggle => {
                        self.stopped = !self.stopped;
                    }
                    CpuControl::Cycle => self.cycle = true,
                    CpuControl::Data => {
                        // self
                        // .state_tx
                        // .send(CpuState {
                        //     a: self.a.clone(),   // Accumulator Register
                        //     x: self.x.clone(),   // X Register
                        //     y: self.y.clone(),   // Y Register
                        //     pc: self.pc.clone(), // Program Counter
                        //     s: self.s.clone(),   // Stack Pointer
                        //     p: self.p.clone(),   // Status Register
                        //     irq: self.irq.clone(),
                        //     nmi: self.nmi.clone(),
                        // })
                        // .unwrap(),
                    }
                }
            }
            None => {}
        }
    }

    pub fn cycle(&mut self) {
        // self.receive_control();

        // if self.stopped & !self.cycle {
        //     self.set_flag(StatusFlag::IrqDisable, true);
        //     return;
        // }
        // self.cycle = false;

        #[cfg(not(target_arch = "wasm32"))]
        while self.pending_cycles != 0 {
            // roughly cycle-accurate timing
            sleep(Duration::from_nanos(500));
            self.pending_cycles -= 1;
        }
        // if !self.get_flag(StatusFlag::IrqDisable) && self.irq {
        //     self.interrupt();
        // }
        let opcode = self.read(self.pc);
        let instruction = get_instruction(opcode);
        instruction.call(self);
    }
    pub fn stop(&mut self) {
        self.stopped = true;
    }

    #[cfg(feature = "debug")]
    pub fn breakpoint(&mut self) {
        self.stop();
    }
}
