use minifb::InputCallback;
use minifb::Key as MKey;

#[cfg(feature = "term")]
use termion::event::Key;

use crate::memory::{MemHandle, MemoryWriter};

#[derive(Debug, Clone)]
pub struct Keyboard {
    memory: MemHandle,
}

impl Keyboard {
    pub fn new(memory: MemHandle) -> Self {
        Self { memory }
    }

    pub fn clear_keys(&self) {
        self.memory.write(0x4400, 0x00);
        self.memory.write(0x4401, 0x00);
        self.memory.write(0x4402, 0x00);
        self.memory.write(0x4403, 0x00);
        self.memory.write(0x4404, 0x00);
        self.memory.write(0x4405, 0x00);
    }

    #[cfg(feature = "term")]
    pub fn read_keys(&self, key: Key) {
        let mut row0 = 0;
        let mut row1 = 0;
        let mut row2 = 0;
        let mut row3 = 0;
        let mut row4 = 0;
        let mut row5 = 0;

        // these are just so the match statements are easier to read lol
        macro_rules! set_shift {
            ($char:expr) => {
                match $char {
                    'A' | 'B' | 'C' | 'D' | 'E' | 'F' | 'G' | 'H' | 'I' | 'J' | 'K' | 'L' | 'M'
                    | 'N' | 'O' | 'P' | 'Q' | 'R' | 'S' | 'T' | 'U' | 'V' | 'W' | 'X' | 'Y'
                    | 'Z' | '!' | '@' | '#' | '$' | '%' | '^' | '&' | '*' | '(' | ')' | '~'
                    | '_' | '+' | '|' | '}' | '{' | '"' | ':' | '?' | '>' | '<' => {
                        row2 ^= 0b1000_0000
                    }
                    _ => {}
                }
            };
        }

        macro_rules! set_plain {
            ($char:expr) => {
                match $char {
                    'w' | 'W' => row0 ^= 0b0100_0000,
                    'e' | 'E' => row0 ^= 0b0010_0000,
                    'r' | 'R' => row0 ^= 0b0001_0000,
                    't' | 'T' => row0 ^= 0b0000_1000,
                    'u' | 'U' => row0 ^= 0b0000_0100,
                    'o' | 'O' => row0 ^= 0b0000_0010,
                    'q' | 'Q' => row1 ^= 0b0100_0000,
                    's' | 'S' => row1 ^= 0b0010_0000,
                    'g' | 'G' => row1 ^= 0b0001_0000,
                    'y' | 'Y' => row1 ^= 0b0000_1000,
                    'i' | 'I' => row1 ^= 0b0000_0100,
                    'p' | 'P' => row1 ^= 0b0000_0010,
                    'd' | 'D' => row2 ^= 0b0100_0000,
                    'v' | 'V' => row2 ^= 0b0010_0000,
                    'h' | 'H' => row2 ^= 0b0001_0000,
                    'k' | 'K' => row2 ^= 0b0000_1000,
                    '\\' | '|' => row2 ^= 0b0000_0100,
                    '/' | '?' => row2 ^= 0b0000_0010,
                    'a' | 'A' => row2 ^= 0b0000_0001,
                    'z' | 'Z' => row3 ^= 0b0100_0000,
                    'f' | 'F' => row3 ^= 0b0010_0000,
                    'b' | 'B' => row3 ^= 0b0001_0000,
                    'j' | 'J' => row3 ^= 0b0000_1000,
                    'l' | 'L' => row3 ^= 0b0000_0100,
                    '2' | '@' => row3 ^= 0b0000_0010,
                    '4' | '$' => row3 ^= 0b0000_0001,
                    'x' | 'X' => row4 ^= 0b0100_0000,
                    'c' | 'C' => row4 ^= 0b0010_0000,
                    'n' | 'N' => row4 ^= 0b0001_0000,
                    'm' | 'M' => row4 ^= 0b0000_1000,
                    ',' | '<' => row4 ^= 0b0000_0100,
                    '1' | '!' => row4 ^= 0b0000_0010,
                    '3' | '#' => row4 ^= 0b0000_0001,
                    ' ' => row5 ^= 0b0100_0000,
                    _ => {}
                }
            };
        }

        match key {
            Key::Esc => row0 ^= 0b1000_0000,
            Key::Backspace => row0 ^= 0b0000_0001,
            Key::Char('\t') => row1 ^= 0b1000_0000,
            Key::Char('\n') | Key::Char('\r') => row1 ^= 0b0000_0001,
            Key::Char(char) => {
                set_shift!(char);
                set_plain!(char);
            }
            Key::ShiftLeft | Key::ShiftUp | Key::ShiftDown | Key::ShiftRight => row2 ^= 0b1000_0000,
            Key::Ctrl(char) => {
                set_shift!(char);
                set_plain!(char);
                row3 ^= 0b1000_0000;
            }
            Key::CtrlUp | Key::CtrlLeft | Key::CtrlDown | Key::CtrlRight => {
                row3 ^= 0b1000_0000;
            }
            Key::Alt(char) => {
                set_plain!(char);
                set_shift!(char);
                row4 ^= 0b1000_0000;
            }
            Key::AltUp | Key::AltLeft | Key::AltDown | Key::AltRight => {
                row4 ^= 0b1000_0000;
            }
            _ => {}
        }

        self.memory.write(0x4400, row0);
        self.memory.write(0x4401, row1);
        self.memory.write(0x4402, row2);
        self.memory.write(0x4403, row3);
        self.memory.write(0x4404, row4);
        self.memory.write(0x4405, row5);

        // keeping this list around to make future changes easier
        // match key.code {
        //     KeyCode::Esc => row0 ^= 0b1000_0000,
        //     KeyCode::Char('w') => row0 ^= 0b0100_0000,
        //     KeyCode::Char('e') => row0 ^= 0b0010_0000,
        //     KeyCode::Char('r') => row0 ^= 0b0001_0000,
        //     KeyCode::Char('t') => row0 ^= 0b0000_1000,
        //     KeyCode::Char('u') => row0 ^= 0b0000_0100,
        //     KeyCode::Char('o') => row0 ^= 0b0000_0010,
        //     KeyCode::Backspace => row0 ^= 0b0000_0001,
        //     KeyCode::Tab => row1 ^= 0b1000_0000,
        //     KeyCode::Char('q') => row1 ^= 0b0100_0000,
        //     KeyCode::Char('s') => row1 ^= 0b0010_0000,
        //     KeyCode::Char('g') => row1 ^= 0b0001_0000,
        //     KeyCode::Char('y') => row1 ^= 0b0000_1000,
        //     KeyCode::Char('i') => row1 ^= 0b0000_0100,
        //     KeyCode::Char('p') => row1 ^= 0b0000_0010,
        //     KeyCode::Enter => row1 ^= 0b0000_0001,
        //     KeyCode::Char('d') => row2 ^= 0b0100_0000,
        //     KeyCode::Char('v') => row2 ^= 0b0010_0000,
        //     KeyCode::Char('h') => row2 ^= 0b0001_0000,
        //     KeyCode::Char('k') => row2 ^= 0b0000_1000,
        //     KeyCode::Char('\'') => row2 ^= 0b0000_0100,
        //     KeyCode::Char('/') => row2 ^= 0b0000_0010,
        //     KeyCode::Char('a') => row2 ^= 0b0000_0001,
        //     KeyCode::Char('z') => row3 ^= 0b0100_0000,
        //     KeyCode::Char('f') => row3 ^= 0b0010_0000,
        //     KeyCode::Char('b') => row3 ^= 0b0001_0000,
        //     KeyCode::Char('j') => row3 ^= 0b0000_1000,
        //     KeyCode::Char('l') => row3 ^= 0b0000_0100,
        //     KeyCode::Char('2') => row3 ^= 0b0000_0010,
        //     KeyCode::Char('4') => row3 ^= 0b0000_0001,
        //     KeyCode::Char('x') => row4 ^= 0b0100_0000,
        //     KeyCode::Char('c') => row4 ^= 0b0010_0000,
        //     KeyCode::Char('n') => row4 ^= 0b0001_0000,
        //     KeyCode::Char('m') => row4 ^= 0b0000_1000,
        //     KeyCode::Char(',') => row4 ^= 0b0000_0100,
        //     KeyCode::Char('1') => row4 ^= 0b0000_0010,
        //     KeyCode::Char('3') => row4 ^= 0b0000_0001,
        //     KeyCode::Char(' ') => row5 ^= 0b0100_0000,
        //     _ => {
        //         row0 = 0;
        //         row1 = 0;
        //         row2 = 0;
        //         row3 = 0;
        //         row4 = 0;
        //         row5 = 0;
        //     }
        // };
    }
}

impl MemoryWriter for Keyboard {
    fn write(&self, address: u16, data: u8) {
        if data != 0x00 {
            // println!("wrote {:02x} to address {:04x}", data, address);
        }
        self.memory.write(address, data);
    }
}

impl InputCallback for Keyboard {
    fn add_char(&mut self, _uni_char: u32) {}
    fn set_key_state(&mut self, key: MKey, _state: bool) {
        let mut row0 = 0;
        let mut row1 = 0;
        let mut row2 = 0;
        let mut row3 = 0;
        let mut row4 = 0;
        let mut row5 = 0;

        match key {
            MKey::Escape => row0 ^= 0b1000_0000,
            MKey::W => row0 ^= 0b0100_0000,
            MKey::E => row0 ^= 0b0010_0000,
            MKey::R => row0 ^= 0b0001_0000,
            MKey::T => row0 ^= 0b0000_1000,
            MKey::U => row0 ^= 0b0000_0100,
            MKey::O => row0 ^= 0b0000_0010,
            MKey::Backspace => row0 ^= 0b0000_0001,
            MKey::Tab => row1 ^= 0b1000_0000,
            MKey::Q => row1 ^= 0b0100_0000,
            MKey::S => row1 ^= 0b0010_0000,
            MKey::G => row1 ^= 0b0001_0000,
            MKey::Y => row1 ^= 0b0000_1000,
            MKey::I => row1 ^= 0b0000_0100,
            MKey::P => row1 ^= 0b0000_0010,
            MKey::Enter => row1 ^= 0b0000_0001,
            MKey::LeftShift | MKey::RightShift => row2 ^= 0b1000_0000,
            MKey::D => row2 ^= 0b0100_0000,
            MKey::V => row2 ^= 0b0010_0000,
            MKey::H => row2 ^= 0b0001_0000,
            MKey::K => row2 ^= 0b0000_1000,
            MKey::Apostrophe => row2 ^= 0b0000_0100,
            MKey::Slash => row2 ^= 0b0000_0010,
            MKey::A => row2 ^= 0b0000_0001,
            MKey::LeftCtrl | MKey::RightCtrl => row3 ^= 0b1000_0000,
            MKey::Z => row3 ^= 0b0100_0000,
            MKey::F => row3 ^= 0b0010_0000,
            MKey::B => row3 ^= 0b0001_0000,
            MKey::J => row3 ^= 0b0000_1000,
            MKey::L => row3 ^= 0b0000_0100,
            MKey::Key2 => row3 ^= 0b0000_0010,
            MKey::Key4 => row3 ^= 0b0000_0001,
            MKey::LeftAlt | MKey::RightAlt => row4 ^= 0b1000_0000,
            MKey::X => row4 ^= 0b0100_0000,
            MKey::C => row4 ^= 0b0010_0000,
            MKey::N => row4 ^= 0b0001_0000,
            MKey::M => row4 ^= 0b0000_1000,
            MKey::Comma => row4 ^= 0b0000_0100,
            MKey::Key1 => row4 ^= 0b0000_0010,
            MKey::Key3 => row4 ^= 0b0000_0001,
            MKey::LeftSuper => row5 ^= 0b1000_0000,
            MKey::Space => row5 ^= 0b0100_0000,
            MKey::RightSuper => row5 ^= 0b0010_0000,
            _ => {}
        };

        self.memory.write(0x4400, row0);
        self.memory.write(0x4401, row1);
        self.memory.write(0x4402, row2);
        self.memory.write(0x4403, row3);
        self.memory.write(0x4404, row4);
        self.memory.write(0x4405, row5);
    }
}
