pub mod cpu;
pub mod error;
pub mod instructions;
pub mod keyboard;
pub mod memory;
pub mod video;

use crate::cpu::Cpu;
use crate::keyboard::Keyboard;
use crate::memory::Mem;
use crate::video::Renderer;

use core::panic;
use cpu::CpuController;
use memory::MemHandle;
use minifb::Window;
use serde::{Deserialize, Serialize};
use std::fmt::Debug;
use std::fs::File;
use std::io::Read;
use std::path::Path;
use std::thread::sleep;
use std::time::{Duration, Instant};

#[cfg(any(feature = "term", feature = "debug"))]
use std::{io::stdout, process::exit};
#[cfg(any(feature = "term", feature = "debug"))]
use termion::{
    async_stdin, clear,
    cursor::{self, Goto},
    event::Key,
    input::TermRead,
    raw::IntoRawMode,
    AsyncReader,
};

#[derive(Serialize, Deserialize, Default, Debug)]
pub enum ScreenType {
    Terminal,
    #[default]
    Window,
}

pub struct Config {
    pub rom: Option<String>,
    pub screen: ScreenType,
    pub char_rom: Option<String>,
}

pub struct GeorgeEmuBuilder<'a> {
    cpu: Cpu,
    renderer: Option<Renderer>,
    input: Option<Keyboard>,
    config: Option<Config>,
    cpu_controller: CpuController,
    memory_handle: MemHandle,
    memory: Mem,
    window: Option<&'a mut Window>,
}

impl<'a> GeorgeEmuBuilder<'a> {
    pub fn new() -> Self {
        let mut memory = Mem::new();
        // TODO: once the memory goes to the handle you can't load a rom,
        // so the method that loads a rom needs to have the handle uninitialized first,
        // so make everything here be None and set everything at the end
        memory.load_bytes(include_bytes!(
            "/Users/august/projects/george-emu/src/roms/demo.rom"
        ));
        let memory_handle = MemHandle::new(memory);
        let cpu_memory = memory_handle.clone();
        let (cpu, cpu_controller) = Cpu::new_with_control(cpu_memory);
        let input = None;
        let config = None;
        let renderer = None;
        let window = None;
        GeorgeEmuBuilder {
            cpu,
            cpu_controller,
            memory,
            memory_handle,
            config,
            renderer,
            window,
            input,
        }
    }

    pub fn config(mut self, config: Config) -> Self {
        self.config = Some(config);
        self
    }

    /// Adds a keyboard input to the emulator
    pub fn keyboard(mut self) -> Self {
        let input_memory = self.memory_handle.clone();
        let keyboard = Keyboard::new(input_memory);
        self.input = Some(keyboard);
        self
    }

    /// Adds a minifb window to the emulator
    #[cfg(not(feature = "term"))]
    pub fn window(mut self, window: &'a mut Window) -> Self {
        let renderer = Renderer::new(self.cpu_controller.clone(), self.memory_handle.clone());
        self.renderer = Some(renderer);
        self.window = Some(window);
        self
    }

    pub fn rom<P>(mut self, path: P) -> Self
    where
        P: AsRef<Path> + Debug + Copy,
    {
        let rom = match File::open(path) {
            Ok(file) => {
                file
                // let mut bin = vec![0; 0x8000];
                // file.read_exact(&mut bin).unwrap();
                // // println!("reading char rom");
                // &bin
            }
            Err(error) => panic!("Couldn't read file at {path:?}: {error:?}"),
        };
        self.memory.read_from_bin(rom).unwrap();
        self
    }

    pub fn build(mut self) -> GeorgeEmu<'a> {
        #[cfg(not(any(feature = "term", feature = "web")))]
        {
            match self.window {
                Some(ref mut window) => {
                    if let Some(ref input) = self.input {
                        window.set_input_callback(Box::new(input.clone()))
                    }
                }
                None => panic!(
                "Tried to call `GeorgeEmuBuilder::build()` before `GeorgeEmuBuilder::window()`!"
            ),
            }
            GeorgeEmu::new(self.cpu, self.renderer.unwrap(), self.input, self.window)
        }
        #[cfg(feature = "term")]
        {
            let renderer = Renderer::new(self.cpu_controller.clone(), self.memory_handle.clone());
            GeorgeEmu::new(self.cpu, renderer, self.input, self.window)
        }
        #[cfg(feature = "web")]
        {
            let renderer = Renderer::new(self.cpu_controller.clone(), self.memory_handle.clone());
            GeorgeEmu::new(self.cpu, renderer, self.input, self.window)
        }
    }
}

#[derive(Debug)]
pub struct GeorgeEmu<'a> {
    pub cpu: Cpu,
    pub renderer: Renderer,
    // in the future these might be set with feature flags,
    // but this is simpler
    input: Option<Keyboard>,
    pub window: Option<&'a mut Window>,
}

impl<'a> GeorgeEmu<'a> {
    pub fn builder() -> GeorgeEmuBuilder<'a> {
        GeorgeEmuBuilder::new()
    }

    fn new(
        mut cpu: Cpu,
        renderer: Renderer,
        input: Option<Keyboard>,
        window: Option<&'a mut Window>,
    ) -> Self {
        cpu.reset();
        Self {
            cpu,
            input,
            renderer,
            window,
        }
    }

    pub fn run(&mut self) {
        #[cfg(any(feature = "term", feature = "debug"))]
        {
            let _ = stdout().into_raw_mode();
            print!("{}{}", cursor::Hide, clear::All,);
        }
        #[cfg(any(feature = "term", feature = "debug"))]
        let mut stdin = async_stdin();
        let clock_interval = Duration::from_nanos(500);
        let frame_interval = Duration::from_millis(16);
        let mut next_clock = Instant::now() + clock_interval;
        let mut next_frame = Instant::now() + frame_interval;

        loop {
            let now = Instant::now();

            if now >= next_clock {
                self.cpu.cycle();
                next_clock += clock_interval;
            }

            if now >= next_frame {
                #[cfg(any(feature = "term", feature = "debug"))]
                self.handle_input(&mut stdin);
                #[cfg(not(target_arch = "wasm32"))]
                self.draw();
                next_frame += frame_interval;
            }

            let next_run = std::cmp::min(next_clock, next_frame);
            let sleep_dur = next_run - now;
            sleep(sleep_dur)
        }
    }

    #[cfg(not(target_arch = "wasm32"))]
    fn draw(&mut self) {
        #[cfg(feature = "term")]
        self.renderer.render();
        #[cfg(not(feature = "term"))]
        match self.window {
            Some(ref mut window) => self.renderer.render(window),
            None => todo!(),
        }
    }
    #[cfg(target_arch = "wasm32")]
    pub fn draw(&self, window: &mut Window) {
        #[cfg(feature = "term")]
        self.renderer.render();
        #[cfg(not(feature = "term"))]
        self.renderer.render(window);
    }

    #[cfg(any(feature = "term", feature = "debug"))]
    fn handle_input(&mut self, stdin: &mut AsyncReader) {
        match &self.input {
            Some(input) => {
                let mut stdin = stdin.keys();
                if let Some(Ok(key)) = stdin.next() {
                    match key {
                        Key::Char('q') => {
                            print!("{}{}{}", clear::All, cursor::Show, Goto(1, 1));
                            exit(0);
                        }
                        Key::Char('`') => self.cpu.toggle_stopped(),
                        Key::Char('\n') => self.cpu.cycle(),
                        Key::Char('i') => self.cpu.interrupt(),
                        _ => {
                            #[cfg(feature = "term")]
                            input.read_keys(key);
                        }
                    }
                } else {
                    #[cfg(feature = "term")]
                    input.clear_keys();
                }
            }
            None => {}
        }
    }

    pub fn cycle(&mut self) {
        self.cpu.cycle()
    }
}
