use anyhow::{bail, Result};
use std::io::{self, Write};
use std::panic;
use std::path::{Path, PathBuf};
use std::str::FromStr;
use std::sync::{Arc, Mutex};
use std::{fs::File, io::Read};

#[derive(Clone, Debug)]
pub struct MemHandle(Arc<Mutex<Mem>>);
impl MemHandle {
    pub fn new(memory: Mem) -> Self {
        Self(Arc::new(Mutex::new(memory)))
    }
    pub fn read(&self, address: u16) -> u8 {
        let memory = self.0.try_lock().unwrap();
        memory.read(address)
    }
    pub fn write(&self, address: u16, data: u8) {
        let mut memory = self.0.try_lock().unwrap();
        memory.write(address, data);
    }

    pub fn dump(&self) {
        let memory = self.0.lock().unwrap();
        let _ = memory.dump(PathBuf::from_str("./cpu_dump.bin").unwrap());
    }
    pub fn poke(&self, address: u16) {
        let memory = self.0.lock().unwrap();
        memory.poke(address)
    }
}

pub trait MemoryReader {
    fn read(&self, address: u16) -> u8;
}

pub trait MemoryWriter {
    fn write(&self, address: u16, data: u8);
}

#[derive(Debug, Clone, Copy)]
// This always feels wrong instead of 0xFFFF,
// but remember it's the number of elements,
// not the maximum index. 0x0000 to 0xFFFF addresses
// 0x10000 elements
pub struct Mem([u8; 0x10000]);

// impl Default for Mem {
//     fn default() -> Self {
//         let bytes = include_bytes!("/Users/august/projects/george-emu/src/roms/george.rom");
//         let padding = [0; 0x8000];
//         let rom: [u8; 0x10000] = {
//             let mut rom: [u8; 0x10000] = [0; 0x10000];
//             let (one, two) = rom.split_at_mut(padding.len());
//             one.copy_from_slice(&padding);
//             two.copy_from_slice(bytes);
//             rom
//         };

//         Self(rom)
//     }
// }

impl Mem {
    pub fn new() -> Self {
        Self([0; 0x10000])
    }
    pub fn dump(&self, path: PathBuf) -> io::Result<()> {
        let mut outfile = File::create(path)?;
        outfile.write_all(&self.0)?;
        Ok(())
    }

    pub fn read(&self, address: u16) -> u8 {
        self.0[address as usize]
    }

    pub fn write(&mut self, address: u16, data: u8) {
        self.0[address as usize] = data;
    }

    pub fn load_rom<P>(&mut self, rom: P) -> Result<()>
    where
        P: AsRef<Path>,
    {
        let rom = match File::open(rom) {
            Ok(rom) => rom,
            Err(_) => bail!("rom could not be opened!"),
        };
        let bytes = rom.bytes();
        for (address, byte) in bytes.enumerate() {
            // println!("{address}");
            match byte {
                Ok(value) => self.write(address as u16 + 0x8000, value),
                Err(_) => {
                    bail!("Loading rom: couldn't write byte {:#04x}", address);
                }
            }
        }
        Ok(())
    }

    pub fn poke(&self, address: u16) {
        println!("{:02x}", self.read(address));
    }
    pub fn load_bytes(&mut self, bytes: &[u8]) {
        for (address, byte) in bytes.into_iter().enumerate() {
            self.write(address as u16 + 0x8000, *byte)
        }
    }

    pub fn read_from_bin(&mut self, f: File) -> Result<()> {
        let bytes = f.bytes();
        for (address, byte) in bytes.enumerate() {
            match byte {
                Ok(value) => self.write(address as u16 + 0x8000, value),
                Err(_) => {
                    bail!("couldn't write byte {:#04x}", address);
                }
            }
        }
        Ok(())
    }
}
